#include "def.h"

.macro carica_descr num base limite pres dpl tipo gran

        #pushl $\gran
        #pushl $\tipo
        #pushl $\dpl
        #pushl $\pres
        #pushl $\limite
        #pushl $\base
        #pushl $\num
        #call  init_descrittore
        #addl $28, %esp

.endm

.text
.global init_descrittore
init_descrittore:
        #pushl %ebp
        #movl %esp, %ebp

        #pushl %ebx
        #pushl %ecx
        #pushl %eax
        #pushl %edx

        #movl $gdt, %ebx
        #movl 8(%ebp), %ecx              // indice GDT -> %ecx
        #movl 16(%ebp), %edx             // limite -> %edx
        #movw %dx,  (%ebx, %ecx, 8)      // bit 15:00 limite -> 1a parola descr.
        #movw 12(%ebp), %ax              // bit 15:00 base -> %ax
        #movw %ax, 2(%ebx, %ecx, 8)      // -> 2a parola descr.
        #movb 14(%ebp), %al              // bit 23:16 base -> %al
        #orb  24(%ebp), %ah              // DPL
        #shlb $5, %ah                    // posizione del DPL nel byte di accesso
        #orb  20(%ebp), %ah              // bit di presenza
        #orb  28(%ebp), %ah              // tipo
        #movw %ax, 4(%ebx, %ecx, 8)      // -> 3a parola descr.
        #movb 15(%ebp), %dh              // bit 31:24 base -> %dh
        #shrl $16, %edx                  // bit 19:16 limite -> low nibble %dl
        #orb  $0b01000000, %dl           // operandi su 32 bit
        #orb  32(%ebp), %dl              // granularita'
        #movw %dx, 6(%ebx, %ecx, 8)      // -> 4a parola descr.
        #
        #popl %edx
        #popl %eax
        #popl %ecx
        #popl %ebx

        #leave
        ret

.global init_gdt
init_gdt:
	pushq %rdx

	movabs $gdt, %rdx
        addq $SEL_SYS_TSS, %rdx
	movw $236, (%rdx) 	//[15:0] = limit[15:0]
	movabsq $sys_tss, %rax
	movw %ax, 2(%rdx)	//[31:16] = base[15:0]
	shrq $16,%rax
	movb %al, 4(%rdx)	//[39:32] = base[24:16]
	movb $0b10001001, 5(%rdx)//[47:40] = p_dpl_type
	movb $0, 6(%rdx)	//[55:48] = 0
	movb %ah, 7(%rdx)	//[63:56] = base[31:24]
	shrq $16, %rax
	movl %eax, 8(%rdx) //[95:64] = base[63:32]
	movl $0, 12(%rdx)	//[127:96] = 0

	popq %rdx
        ret
        

.global load_gdt
load_gdt:
        lgdt    gdt_pointer             // carichiamo la nuova GDT

        movw $SEL_SYS_TSS, %cx
        ltr %cx
        
        retq

.global unload_gdt
unload_gdt:
        #pushl %ecx

        #lgdt  save_gdt
        #
        #pushf
        #xorl %ecx, %ecx
        #movw  s_CS, %cx
        #pushl %ecx
        #pushl $avanti
        #iret
avanti:
        #movw    s_DS, %cx
        #movw    %cx, %ds
        #movw    s_ES, %cx
        #movw    %cx, %es
        #movw    s_FS, %cx
        #movw    %cx, %fs
        #movw    s_GS, %cx
        #movw    %cx, %gs
        #movw    s_SS, %cx
        #movw    %cx, %ss

	#popl %ecx
        ret

// carica un gate nella IDT
// parametri: 
// num: indice (a partire da 0) in IDT del gate da caricare
// routine: indirizzo della routine da associare al gate
.macro carica_gate num routine

	movabsq $(idt + \num*16), %rdi
	movq $\routine, %rsi
	call componi_gate

.endm

.global componi_gate
componi_gate:
	movq %rsi, %rax		// offset della routine

	movw %ax, (%rdi)  	// primi 16 bit dell'offset
	movw $SEL_CODICE_SISTEMA, 2(%rdi)

	movw $0, %ax
	movb $0b10001110, %ah 	        // byte di accesso
					// (presente, 32bit, tipo interrupt)
	movb $0, %al			// la parte bassa deve essere 0
	movl %eax, 4(%rdi)	// 16 bit piu' sign. dell'offset
					// e byte di accesso
	shrq $32, %rax			//estensione a 64 bit dell'offset
	movl %eax, 8(%rdi)
	movl $0, 12(%rdi) 		//riservato

	ret

.global init_idt
init_idt:
        pushq %rcx
        pushq %rax
        pushq %rdi

        movabsq $idt, %rdi
        movq $256, %rcx
        movq $0, %rax
        cld
        rep stosq

	carica_gate	0 		divide_error 	
	carica_gate	1 		debug 		
	carica_gate	2 		nmi 		
	carica_gate	3 		breakpoint 	
	carica_gate	4 		overflow 	
	carica_gate	5 		bound_re 	
	carica_gate	6 		invalid_opcode	
	carica_gate	7 		dev_na 		
	carica_gate	8 		double_fault 	
	carica_gate	9 		coproc_so 	
	carica_gate	10 		invalid_tss 	
	carica_gate	11 		segm_fault 	
	carica_gate	12 		stack_fault 	
	carica_gate	13 		prot_fault 	
	carica_gate	14 		int_tipo_pf 	
	carica_gate	15		unknown_exc	
	carica_gate	16 		fp_exc 		
	carica_gate	17 		ac_exc 		
	carica_gate	18 		mc_exc 		
	carica_gate	19 		simd_exc 	
	carica_gate	0x20 		ignore_pic 	
	carica_gate	0x21 		ignore_pic 	
	carica_gate	0x22 		ignore_pic 	
	carica_gate	0x23 		ignore_pic 	
	carica_gate	0x24 		ignore_pic 	
	carica_gate	0x25 		ignore_pic 	
	carica_gate	0x26 		ignore_pic 	
	carica_gate	0x27 		ignore_pic 	
	carica_gate	0x28 		ignore_pic 	
	carica_gate	0x29 		ignore_pic 	
	carica_gate	0x2A 		ignore_pic 	
	carica_gate	0x2B 		ignore_pic 	
	carica_gate	0x2C 		ignore_pic 	
	carica_gate	0x2D 		ignore_pic 	
	carica_gate	0x2E 		ignore_pic 	
	carica_gate	0x2F 		ignore_pic 	

        lidt    idt_pointer                // carichiamo la nuova IDT

        popq %rdi
        popq %rax
        popq %rcx
        ret

divide_error:
	movq $0, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

debug:
	movq $1, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

nmi:
	movq $2, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

breakpoint:
	movq $3, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

overflow:
	movq $4, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

bound_re:
	movq $5, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

invalid_opcode:
	movq $6, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

dev_na:
	movq $7, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

double_fault:
	movq $8, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

coproc_so:
	movq $9, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

invalid_tss:
	movq $10, %rdi
	popq %rdi
	popq %rdx
	jmp comm_exc

segm_fault:
	movq $11, %rdi
	popq %rdi
	popq %rdx
	jmp comm_exc

stack_fault:
	movq $12, %rdi
	popq %rdi
	popq %rdx
	jmp comm_exc

prot_fault:
	movq $13, %rdi
	popq %rdi
	popq %rdx
	jmp comm_exc

int_tipo_pf:
	movq $14, %rdi
	popq %rsi
	popq %rdx
	jmp comm_exc

unknown_exc:
	movq $15, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

fp_exc:
	movq $16, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

ac_exc:
	movq $17, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

mc_exc:
	movq $18, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc

simd_exc:
	movq $19, %rdi
	movq $0, %rsi
	popq %rdx
	jmp comm_exc


comm_exc:
	call gestore_eccezioni
	hlt

ignore_pic:
	call apic_send_EOI
	iret

.global idt_reset
idt_reset:
    #lidt    save_idt
    ret

.global sti
sti:   sti
        ret

.global cli
cli:   cli
        ret

.data
.balign 16
.global idt
idt:
        .space 16 * 256, 0
sys_tss:
        .space 236, 0
.global gdt
gdt:
	.quad 0		//segmento nullo
code_sys_seg:
	.word 0b0           //limit[15:0]   not used
	.word 0b0           //base[15:0]    not used
	.byte 0b0           //base[23:16]   not used
	.byte 0b10011010    //P|DPL|1|1|C|R|A|  DPL=00=sistema
	.byte 0b00100000    //G|D|L|-|-------|  L=1 long mode
	.byte 0b0           //base[31:24]   not used
code_usr_seg:
	.word 0b0           //limit[15:0]   not used
	.word 0b0           //base[15:0]    not used
	.byte 0b0           //base[23:16]   not used
	.byte 0b11111010    //P|DPL|1|1|C|R|A|  DPL=11=utente
	.byte 0b00100000    //G|D|L|-|-------|  L=1 long mode
	.byte 0b0           //base[31:24]   not used
data_usr_seg:
	.word 0b0           //limit[15:0]   not used
	.word 0b0           //base[15:0]    not used
	.byte 0b0           //base[23:16]   not used
	.byte 0b11110010    //P|DPL|1|0|E|W|A|  DPL=11=utente
	.byte 0b00000000    //G|D|-|-|-------|  
	.byte 0b0           //base[31:24]   not used
des_tss:
	.quad 0
	.quad 0
end_gdt:

gdt_pointer:
        .word end_gdt-gdt               // limite della GDT
        .quad gdt                       // base della GDT
idt_pointer:
        .word 0xFFF                     // limite della IDT (256 entrate)
        .quad idt                       // base della IDT
